"""Class that encapsulate the docker client and provide common methods"""
import atexit
import logging
import os
import re
import tarfile
from functools import partial
from multiprocessing import Process
from typing import Any, Dict, List, Mapping, Optional, Pattern, Tuple

import docker  # type: ignore[import]
import dockerpty  # type: ignore[import]
from docker.models.containers import Container  # type: ignore[import]
from docker.types import Mount  # type: ignore[import]

from .config import ODOO_7, ODOO_8, ODOO_9, ODOO_10, Configuration

_logger = logging.getLogger(__name__)

# TODO detect that user is member of docker group


class DockerClient:
    """Provide additional abstraction and extra methods of the docker client API"""

    client = docker.from_env()

    @classmethod
    def create_volume(
        cls, volume_name: str, extra_labels: Optional[Dict[str, str]] = None
    ) -> Tuple[docker.models.volumes.Volume, bool]:
        """Return the volume passed in parameter, creating it if it does not exist.
        :param volume_name: name of the volume to create
        :param extra_labels: extra labels to put on the volume (only on
        creation)
        :return: a tuple, the first one is the volume, the second is a boolean
        indicating that the volume has just been created (true if created)
        """
        volumes = cls.client.volumes.list(filters={"name": volume_name})
        if volumes:
            _logger.debug("Volume %s already exists", volume_name)
            return volumes[0], False
        _logger.debug("Creating volume %s", volume_name)
        labels = dict(odoo_scripts="")
        if extra_labels:
            labels.update(extra_labels)
        return cls.client.volumes.create(name=volume_name, labels=labels), True

    @staticmethod
    def put_file(src, container, dst):
        """put a file into a container"""
        # copied from https://stackoverflow.com/questions/46390309/how-to-copy-a-file-from-host-to-container-using-docker-py-docker-sdk # noqa: E501 # pylint: disable=line-too-long
        os.chdir(os.path.dirname(src))
        srcname = os.path.basename(src)
        # TODO put that in a temporary directory
        tar = tarfile.open(src + ".tar", mode="w")
        try:
            tar.add(srcname)
        finally:
            tar.close()

        data = open(src + ".tar", "rb").read()
        container.put_archive(os.path.dirname(dst), data)

    @classmethod
    def exec(
        cls,
        container: docker.models.containers.Container,
        command: List[str],
        user: str,
        interactive: bool = False,
    ) -> int:
        """Execute a command in a container

        :param container: Container to exec on
        :param command: Command to be executed
        :param user: User to execute command as.
        :param interactive: True if the container needs to be interactive
        """
        if interactive:
            dockerpty.exec_command(cls.client.api, container.id, command)
            # it would be nice to return the exec command return code, but
            # there is no way to do that apparently
            result = 0
        else:
            run_result = container.exec_run(command, user=user)
            result = run_result.exit_code
            print(run_result.output.decode().rstrip())
        return result

    @staticmethod
    def _replace_then_print(replaces: List[Tuple[Pattern[str], str]], log_line: str):
        # apply all the replacement of the path inside the image to the real path
        for pattern, value in replaces:
            log_line = pattern.sub(value, log_line)
        # print to play nicely with pycharm that will linkify the result
        print(log_line)

    @classmethod
    def run(
        cls,
        repository: str,
        tag: str,
        run_kwargs: Dict[str, Any],
        pull: bool = True,
        target_source_dict: Optional[Mapping[str, str]] = None,
        display_log: bool = True,
    ):
        """Run a container with the provided args"""
        replaces: List[Tuple[Pattern[str], str]] = (
            [(re.compile(key), value) for key, value in target_source_dict.items()]
            if target_source_dict
            else []
        )

        image = f"{repository}:{tag}"

        if pull:
            _logger.debug("Pulling %s", image)
            # TODO find a better way to avoid error when pulling without a connection
            try:
                cls.client.images.pull(repository, tag)
            except docker.errors.APIError:
                _logger.warning(
                    "No connection to server, using existing version, if any"
                )

        container = None

        # XXX When pressing ctrl-c during the container creating, the container
        #  is not removed as there is no return of the value of the Container
        #  object. There is several ways to do that:
        #  - give the container a unique name we know, and use that to clean
        #  - maybe use some kind of time frame to identify the container
        #  - warn the user of a potential container leak by using signals

        def remove():
            """Remove the container"""
            if container:
                _logger.debug("Removing container %s", container.name)
                container.remove()

        def stop_remove():
            """Stop then remove the container."""
            if container:
                _logger.debug("Stopping container %s", container.name)
                container.stop()
                _logger.debug("Waiting for container %s", container.name)
                container.wait()
            remove()

        atexit.register(stop_remove)

        # split create and start so in case of SIGTERM, container has fewer chances
        # of being None
        kwargs = dict(run_kwargs, detach=True)
        _logger.debug("Create container with image %s", image)
        container = cls.client.containers.create(image, **kwargs)
        _logger.debug("Created container %s", container.name)
        _logger.debug("Starting container %s", container.name)
        container.start()
        if run_kwargs.get("detach", False):
            if display_log:
                process = Process(target=DockerClient.print_log, args=(container,))
                process.start()

                def stop_remove_wait_process():
                    stop_remove()
                    if process and process.is_alive():
                        _logger.info("Waiting for log process to finish")
                        process.join()

                atexit.register(stop_remove_wait_process)
                atexit.unregister(stop_remove)

            return container

        atexit.unregister(stop_remove)
        atexit.register(remove)

        if kwargs.get("tty", False):
            line = b""
            for log in container.logs(stream=True, follow=True):
                # if there is a tty, send byte by byte, so add them together before
                # printing
                # works correctly because prettier sends \r\n on each line
                if log == b"\r":
                    cls._replace_then_print(replaces, line.decode())
                    line = b""
                elif log == b"\n":
                    pass
                else:
                    line += log
            if line:
                cls._replace_then_print(replaces, line.decode())
        else:
            cls.print_log(container, replaces)

        _logger.debug("Waiting for container %s", container.name)
        result = container.wait()

        if result.get("Error", None):
            _logger.warning(result.decode())
        _logger.debug("Removing container %s", container.name)
        container.remove(v=True)
        atexit.unregister(remove)
        return result["StatusCode"]

    @classmethod
    def print_log(
        cls,
        container: Container,
        replaces: Optional[List[Tuple[Pattern[str], str]]] = None,
    ):
        """Print the logs of a container, eventually replacing some of its output."""
        if replaces is None:
            replaces = []
        for log in container.logs(stream=True, follow=True):
            # \r is needed when using a process that also outputs the logs
            log = log.decode().rstrip() + "\r"
            cls._replace_then_print(replaces, log)


def get_odoo_base_path(config: Configuration) -> str:
    """Return value of env ODOO_BASE_PATH in the local image name."""
    image = DockerClient.client.images.list(config.local_image_name)
    if not image:
        raise Exception(
            "Image %s is not built, needed to be inspected", config.local_image_name
        )
    for env in image[0].attrs["Config"]["Env"]:
        if env.startswith("ODOO_BASE_PATH="):
            return env[15:]
    raise Exception("Missing ODOO_BASE_PATH in the image")


def modules_mount(config: Configuration) -> Tuple[List[Mount], Dict[str, str]]:
    """Return a list of Mount for the modules in the configuration"""
    modules = config.modules
    _logger.debug("addon modules: %s", ",".join(config.modules))
    project_path = config.path
    mount_list: List[Mount] = []
    mount_dict: Dict[str, str] = {}
    if config.odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        base_python_path = "/usr/local/lib/python2.7/dist-packages/"
        target_path = "/opt/odoo/sources/odoo/addons"
    else:
        base_python_path = os.path.dirname(get_odoo_base_path(config))
        target_path = get_odoo_base_path(config) + "/addons"

    for module in modules:
        full_target_path = os.path.join(target_path, os.path.basename(module))
        full_project_path = os.path.realpath(os.path.join(project_path, module))
        mount_list.append(Mount(full_target_path, full_project_path, "bind"))
        mount_dict[full_target_path] = full_project_path

    _logger.debug("python packages: %s", ",".join(config.python_packages))
    for python_package in config.python_packages.items():
        for dir_ in python_package[1]["dirs"]:
            full_target_path = os.path.join(base_python_path, dir_[1])
            full_project_path = os.path.realpath(
                os.path.join(project_path, python_package[0], dir_[0])
            )
            mount_list.append(Mount(full_target_path, full_project_path, "bind"))
            mount_dict[full_target_path] = full_project_path

    return mount_list, mount_dict


def anthem_mounts(config: Configuration) -> List[Mount]:
    """Return a list of mounts for all the anthem songs present in the super project."""
    project_path = os.path.realpath(config.path)
    mounts: List[Mount] = []
    # No anthem on old ODOO
    if config.odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
        return mounts
    # Mount songs for anthem to avoid having to rebuild the image when they are changed.
    # Assume they are in songs/something and put in the image in the same place Odoo is
    # installed.
    if os.path.exists("songs"):
        for root, dirs, _files in os.walk("songs"):
            for directory in dirs:
                mounts.append(
                    Mount(
                        os.path.dirname(get_odoo_base_path(config))
                        + f"/{root}/{directory}",
                        os.path.join(project_path, root, directory),
                        "bind",
                        read_only=True,
                    )
                )
    _logger.debug("mounts len: %s", len(mounts))
    return mounts


def odoo_source_mounts(config: Configuration, source_path: str) -> List[Mount]:
    """Return a list of mounts for a directory containing Odoo sources."""
    mounts: List[Mount] = []
    full_source_path = os.path.realpath(source_path)
    if config.odoo_type in (ODOO_7, ODOO_8):
        mounts.append(Mount("/opt/odoo/sources/odoo", source_path, "bind"))
    else:
        install_path = (
            "/usr/local/lib/python2.7/dist-packages/odoo"
            if config.odoo_type in (ODOO_9, ODOO_10)
            else get_odoo_base_path(config)
        )
        for entry in os.listdir(os.path.join(full_source_path, "odoo")):
            if entry != "addons":
                mounts.append(
                    Mount(
                        install_path + "/" + entry,
                        os.path.join(full_source_path, "odoo", entry),
                        "bind",
                    )
                )
        for entry in os.listdir(os.path.join(full_source_path, "odoo", "addons")):
            # for directory in dirs:
            mounts.append(
                Mount(
                    install_path + "/addons/" + entry,
                    os.path.join(full_source_path, "odoo", "addons", entry),
                    "bind",
                )
            )
        # iterate on all the modules, avoids
        for entry in os.listdir(os.path.join(full_source_path, "addons")):
            mounts.append(
                Mount(
                    install_path + "/addons/" + entry,
                    os.path.join(full_source_path, "addons", entry),
                    "bind",
                )
            )
    return mounts


class DockerService:
    """Represent a service in a docker and allows common methods to manipulate it."""

    def __init__(
        self,
        docker_client: docker.client.DockerClient,
        name: str,
        repository: str,
        tag: str,
        stop_remove_if_exists: bool,
        immediate_pull: bool = False,
        attach: bool = True,
        stream: bool = True,
        stop_at_exit: bool = True,
        force_pull: bool = False,
    ):
        """
        :param docker_client: instance of docker client
        :param name: name for the container
        :param repository: repository to use for the image
        :param tag: repository tag to use for the image
        :param immediate_pull: pull now (rather than do it later, like during start)
        :param stream: if not attached, indicate to stream output or not
        """
        super().__init__()
        self.docker_cli = docker_client
        self.name = name
        self.repository: str = repository
        self.tag = tag
        self.base_image = f"{self.repository}:{self.tag}"
        if immediate_pull:
            self.pull_image()
        self.attach = attach
        self.stream = stream
        self.stop_at_exit = stop_at_exit
        # Retrieve the container if it already exists
        self._container: Optional[docker.models.containers.Container]
        try:
            self._container = self.docker_cli.containers.get(self.name)
        except docker.errors.NotFound:
            self._container = None
        self._log_process: Optional[Process] = None
        self.stop_remove_if_exists = stop_remove_if_exists
        self.force_pull: bool = force_pull

    def start(self) -> bool:
        """Start the container"""
        # TODO avoid pull if already done with immediate pull
        self.pull_image()
        if self._container is not None:
            # TODO should it crash?
            _logger.info("Container %s already exist", self.name)
            if self.stop_remove_if_exists:
                # stop and remove
                self.stop_and_remove()
            elif not self.is_running:
                _logger.info("Container %s is not running", self.name)
                self.remove()
            else:
                return True
        self._start()
        if self.stop_at_exit:
            atexit.register(partial(self.stop_and_remove, False))
        if self.attach:
            DockerClient.print_log(self._container)
        elif self.stream:
            process = Process(target=DockerClient.print_log, args=(self._container,))
            process.start()
            self._log_process = process

        return True

    def _create_param(self) -> dict:
        return dict(
            name=self.name,
            network_mode="host",
            image=self.base_image,
        )

    def _start(self) -> None:
        self._container = self.docker_cli.containers.create(**self._create_param())
        _logger.info("Starting container %s", self.name)
        self._container.start()

    def stop_and_remove(self, expect_running: bool = True) -> None:
        """Stop and remove the container"""
        try:
            self.stop(expect_running)  # call stop func
            self.remove()  # call remove func
        except docker.errors.NotFound as exception:
            _logger.info("Stop and remove container, got: %s", exception)
            raise exception

    def stop(self, expect_running: bool = True) -> None:
        """Stop service container."""
        try:
            if not self.is_running:
                if expect_running:
                    _logger.warning("Container %s is not running", self.name)
            elif self._container is not None:
                _logger.info("Stopping %s container", self.name)
                self._container.stop()
                self._container.wait()
        except docker.errors.NotFound as exception:
            _logger.error("Stopping container %s, got: %s", self.name, exception)
            raise
        self.join_log_process()

    def stop_log_process(self):
        if self._log_process and self._log_process.is_alive():
            self._log_process.terminate()

    def join_log_process(self):
        if self._log_process and self._log_process.is_alive():
            _logger.info("Waiting for log process to finish")
            self._log_process.join()

    def remove(self):
        """Remove the service container created"""
        try:
            if self._container is not None:
                _logger.info("Removing %s container", self.name)
                self._container.remove()
            else:
                _logger.info("Container %s already removed", self.name)
        except docker.errors.NotFound as exception:
            _logger.error("Removing container %s, got: %s", self.name, exception)
            raise exception

    def pull_image(self):
        """Pulls image if needed or forced."""
        if self.force_pull or not self.docker_cli.images.list(self.base_image):
            _logger.info("Pulling %s", self.base_image)
            self.docker_cli.images.pull(repository=self.repository, tag=self.tag)

    @property
    def is_running(self) -> bool:
        """Verify the container status
        :return: true if the status is running
        """
        if self._container is None:
            return False
        result = self.docker_cli.containers.list(filters={"id": self._container.id})
        return result and result[0].status == "running"

    @property
    def is_exited(self) -> bool:
        """
        Verify the rednerd container status
        :return: Boolean if the status is exited
        """
        if self._container is None:
            return True
        result = self.docker_cli.containers.list(filters={"id": self._container.id})
        return result and result[0].status == "exited"


def get_volumes(base_dir, directory: str, write: bool = False):
    """return volumes to bind"""
    path = os.path.abspath(os.path.join(base_dir, directory))
    volumes = {
        base_dir: {
            "bind": base_dir,
            "mode": "rw" if write else "ro",
        }
    }
    # also bind any symbolic link, but only for the given directory or its parent.
    # all symbolic link could be scanned for but that might take too much time
    for potential_link in (path, os.path.dirname(path)):
        if os.path.islink(potential_link):
            volumes[potential_link] = {
                "bind": os.path.realpath(potential_link),
                "mode": "rw" if write else "ro",
            }
    return volumes, path
