#!/usr/bin/env python3
"""Launch a docker image but bind the local directory inside the container
and define the module path automatically.
"""
import atexit
import logging
import os
import pwd
import socket
import sys
import tempfile
from argparse import ArgumentParser
from configparser import ConfigParser
from contextlib import closing
from subprocess import call, check_output
from typing import Any, Dict, List, Optional

import dockerpty  # type: ignore[import]
from docker.errors import APIError  # type: ignore[import]
from docker.types import Mount  # type: ignore[import]
from psycopg2 import OperationalError, connect
from psycopg2.errors import DuplicateDatabase, DuplicateObject

from .config import (
    ODOO_7,
    ODOO_8,
    ODOO_9,
    ODOO_10,
    ODOO_11,
    ODOO_12,
    ODOO_13,
    ODOO_14,
    ODOO_15,
    ODOO_16,
    Config,
)
from .docker_build import add_build_options, build_local_image, get_build_options
from .docker_client import (
    DockerClient,
    anthem_mounts,
    modules_mount,
    odoo_source_mounts,
)
from .docker_flake8 import apply_flake8, parser_add_flake8_group
from .docker_isort import apply_isort, parser_add_isort_group
from .docker_postgresql import POSTGRES_PASSWORD, docker_run_postgresql
from .docker_py3o import start_py3o
from .docker_pylint import apply_pylint, parser_add_pylint_group
from .docker_redis import DockerRedis
from .docker_rednerd import DockerRednerd
from .parsing import apply, basic_parser

# TODO auto create list of module

_logger = logging.getLogger(__name__)

__version__ = "3.14.0"
__date__ = "2017-08-11"
__updated__ = "2023-07-21"


def __parser(project_name: str) -> ArgumentParser:
    """Return a parser for docker_dev_start"""
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = f"""{program_shortdesc}

  Created by Vincent Hatakeyama on {__date__}.
  Copyright 2017-2023 XCG Consulting. All rights reserved.

  Licensed under the MIT License

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
"""

    # TODO add a way to store configuration options in a project file
    # Argument parsing
    parser = basic_parser(description=program_license, version=program_version_message)
    parser.add_argument(
        "--name",
        help="Container name to use [default: %(default)s]",
        default=project_name,
    )
    parser.add_argument(
        "--db_user", help="Database user [default: %(default)s]", default=None
    )
    parser.add_argument(
        "--db_password",
        help="Database user password [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--db_host",
        help="database host address",
    )
    parser.add_argument(
        "--max-cron-threads",
        help="max cron threads [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "-d",
        "--database",
        help="Database [default: %(default)s]",
        default=None,
    )
    parser.add_argument("-p", "--dbport", help="Change database port", default=None)
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-u",
        "--update",
        help="Module to update (will also set --i18n-overwrite)"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "--install-test",
        help="Modules to install for tests (will also set --log-level=test)"
        " [default: %(default)s]\n"
        "Options --update, --test, --install-test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "-t",
        "--test",
        help="Modules to test (will also set --log-level=test)"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        default=None,
    )
    group.add_argument(
        "--test-default",
        help="Test all modules in module_list_test"
        " [default: %(default)s]\n"
        "Options --update, --test and --test-default"
        " cannot be used at the same time",
        action="store_true",
    )
    parser.add_argument(
        "--test-tags",
        help="Option sent to odoo [default: %(default)s]",
        default=None,
    )
    populate = group.add_argument_group()
    populate.add_argument(
        "--populate-model",
        help="Populate the indicated model",
        default=None,
    )
    populate.add_argument(
        "--populate-size",
        help="Populate size",
        default=None,
    )
    parser.add_argument(
        "-i",
        "--install",
        help="Module to install [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--without-demo",
        help="Module to avoid including demo data [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--create-user",
        help="Create user in database (using $user) [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--install-default",
        help=(
            "Append the module_list from setup.cfg to the modules to install"
            " [default: %(default)s]"
        ),
        action="store_true",
    )
    parser.add_argument(
        "--start-postgresql",
        help="Start a Postgresql docker",
        action="store_true",
    )
    parser_add_flake8_group(parser)
    parser_add_isort_group(parser)
    parser_add_pylint_group(parser)
    add_build_options(parser)
    parser.add_argument(
        "--odoo-help",
        help="Pass --help to odoo binary [default: %(default)s]",
        action="store_true",
    )
    parser.add_argument(
        "--stop-after-init",
        help="Pass --stop-after-init to odoo binary [default: %(default)s]",
        action="store_true",
    )
    build_group = parser.add_mutually_exclusive_group()
    build_group.add_argument(
        "--no-build",
        help="Do not build locally [default: %(default)s]",
        action="store_true",
    )
    build_group.add_argument(
        "--force-build",
        help="Force new local build [default: %(default)s]",
        action="store_true",
    )
    network_group = parser.add_mutually_exclusive_group()
    network_group.add_argument(
        "--host-network",
        help="Use host network [default: %(default)s]",
        action="store_true",
        default=True,
    )
    network_group.add_argument(
        "--docker-network",
        help=(
            "Use docker network (works better with"
            " python-netifaces installed) [default: %(default)s]"
        ),
        action="store_false",
        dest="host_network",
        default=False,
    )
    dev_option_group = parser.add_mutually_exclusive_group()
    dev_option_group.add_argument(
        "--no-dev", help="No dev options", action="store_false", dest="dev"
    )
    dev_option_group.add_argument(
        "--dev-opt",
        help="add extra dev options",
        nargs="+",
        dest="dev_opts",
        default=[],
    )
    parser.add_argument(
        "--load-language",
        help="specifies the languages for the translations you want to be"
        " loaded (also read from configuration file)",
        default=None,
        dest="LOAD_LANGUAGE",
    )
    # TODO there are more
    odoo_log_levels = ["info", "warn", "debug", "debug_sql"]
    parser.add_argument(
        "--log-level",
        help="Override odoo log level (for tests and install)",
        default=None,
        choices=odoo_log_levels,
    )
    parser.add_argument(
        "--log-handler",
        help="log handler [default: %(default)s]",
        default=None,
        action="append",
    )
    parser.add_argument(
        "--restore",
        help="Database to restore [default: %(default)s]",
        default=None,
        dest="DUMP",
    )
    parser.add_argument(
        "--config",
        help="Configuration file to use, if any [default: %(default)s]",
        default=os.path.join("conf", "dev", "odoo.conf"),
    )

    py3o_group = parser.add_mutually_exclusive_group()
    py3o_group.add_argument(
        "--start-py3o",
        help="start py3o docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_py3o",
    )
    py3o_group.add_argument(
        "--no-start-py3o",
        help="do not start py3o docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_py3o",
    )
    rednerd_group = parser.add_mutually_exclusive_group()
    rednerd_group.add_argument(
        "--start-rednerd",
        help="start rednerd docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_rednerd",
    )
    rednerd_group.add_argument(
        "--no-start-rednerd",
        help="do not start rednerd docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_rednerd",
    )
    redner_options_group = parser.add_argument_group()
    redner_options_group.add_argument(
        "--redner-db-user",
        help="Database user for redner [default: %(default)s]",
        default="redner",
    )
    redner_options_group.add_argument(
        "--redner-db-password",
        help="Database password for redner [default: %(default)s]",
        default="redner",
    )
    redner_options_group.add_argument(
        "--redner-admin-password",
        help="Redner admin password. Will be updated in redner with this value "
        "[default: %(default)s]",
        default="admin",
    )
    redner_options_group.add_argument(
        "--redner-user",
        help="Redner user that will be used. Created if it does not exists [default: "
        "%(default)s]",
        default="odoo",
    )
    redner_options_group.add_argument(
        "--redner-password",
        help="Redner user that will be used. Will be updated in redner [default: "
        "%(default)s]",
        default="redner-odoo-password",
    )
    redis_group = parser.add_mutually_exclusive_group()
    redis_group.add_argument(
        "--start-redis",
        help="start redis docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=1,
        dest="start_redis",
    )
    redis_group.add_argument(
        "--no-start-redis",
        help="do not start redis docker [default: %(default)s]",
        action="store_const",
        default=None,
        const=-1,
        dest="start_redis",
    )

    parser.add_argument(
        "--odoo-sources",
        help="Odoo sources to use, if any [default: %(default)s]",
        default=None,
    )
    parser.add_argument(
        "--pre-sql-script",
        help="SQL script to run on database [default: %(default)s]",
        default=None,
        dest="PRE_SQL",
    )

    parser.add_argument(
        "--chown",
        help="run chown [default: %(default)s]",
        action="store_true",
        dest="run_chown",
    )

    marabunta_group = parser.add_mutually_exclusive_group()
    marabunta_group.add_argument(
        "--no-marabunta",
        help="Do not run marabunta",
        action="store_false",
        default=True,
        dest="marabunta",
    )
    marabunta_options_group = marabunta_group.add_argument_group()
    marabunta_options_group.add_argument(
        "--marabunta-db-user",
        help="specify a db user for marabunta, this is useful if the odoo user is not "
        "able to create databases or extensions",
    )
    marabunta_options_group.add_argument(
        "--marabunta-db-password",
        help="specify a db password for marabunta",
    )
    marabunta_options_group.add_argument(
        "--marabunta-force-version",
        help="marabunta force version argument (see marabunta documentation)",
    )

    pythondevmode_group = parser.add_mutually_exclusive_group()
    pythondevmode_group.add_argument(
        "--pythondevmode",
        help="Set PYTHONDEVMODE",
        action="store_true",
        default=False,
        dest="force_python_dev_mode",
    )
    pythondevmode_group.add_argument(
        "--no-pythondevmode",
        help="Do not set PYTHONDEVMODE",
        action="store_true",
        default=False,
        dest="force_no_python_dev_mode",
    )

    # workers options
    parser.add_argument(
        "--workers",
        help="Number of workers [default: %(default)s]",
        default=0,
        type=int,
    )

    # workers options
    parser.add_argument(
        "--activate-s3-storage",
        help="Activate S3 storage using the provided bucket as the main one [default: "
        "%(default)s]",
        default=None,
        type=str,
    )

    return parser


def set_rednerd_configparameters(cursor, user, api_key, server_url) -> None:
    for key, value in (
        ("redner.account", user),
        ("redner.api_key", api_key),
        ("redner.server_url", server_url),
    ):
        if value is not None:
            cursor.execute(
                "INSERT INTO ir_config_parameter (key, value) VALUES (%s, %s) ON "
                "CONFLICT (key) DO UPDATE SET value=%s WHERE ir_config_parameter.key = "
                "%s",
                (key, value, value, key),
            )


def main(argv: Optional[List[str]] = None) -> int:
    """Parse arguments and launch conversion"""
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)

    # basic detection to avoid errors
    if project_name == "odoo_scripts":
        logging.fatal(
            "You must run this script from the super project"
            " (./odoo_scripts/docker_dev_start.py)"
        )
        return 1

    # TODO detect that user is member of docker group

    parser = __parser(project_name)

    # TODO add a way to add options to docker
    # TODO add a way to add arg to odoo

    nmspc = parser.parse_args(argv)
    project_name = nmspc.name
    apply(nmspc)
    db_user = nmspc.db_user
    db_password = nmspc.db_password
    use_host_network = nmspc.host_network
    _logger.debug("Use host network: %s", str(use_host_network))
    start_postgresql = nmspc.start_postgresql
    odoo_help = nmspc.odoo_help
    dev = nmspc.dev
    dev_opts = nmspc.dev_opts
    restore_filename = nmspc.DUMP
    database = nmspc.database
    """Odoo database. Optional[str]"""
    load_language = ""
    installing_or_updating = False
    odoo_sources = nmspc.odoo_sources
    pre_sql_script = nmspc.PRE_SQL
    run_chown = nmspc.run_chown
    populate_model = nmspc.populate_model
    populate_size = nmspc.populate_size
    marabunta = nmspc.marabunta
    marabunta_db_user = nmspc.marabunta_db_user
    marabunta_db_password = nmspc.marabunta_db_password
    marabunta_force_version = nmspc.marabunta_force_version
    force_python_dev_mode = nmspc.force_python_dev_mode
    force_no_python_dev_mode = nmspc.force_no_python_dev_mode
    workers = nmspc.workers
    gevent = workers > 0
    test_tags = nmspc.test_tags
    activate_s3_storage = nmspc.activate_s3_storage

    if restore_filename:
        if not database:
            _logger.fatal("No database name given for restore")
            return 13

    config = Config()

    if not db_user and config.db_user is not None:
        db_user = config.db_user
    if not db_password and config.db_password is not None:
        db_password = config.db_password
    if config.load_language is not None:
        load_language = config.load_language

    odoo_type = config.odoo_type
    image = config.local_image_name
    postgresql_version = config.postgresql_version
    module_list = config.module_list
    if nmspc.start_py3o:
        _logger.debug("Command line start py3o %d", nmspc.start_py3o)
        start_py3o_stack = nmspc.start_py3o == 1
    else:
        start_py3o_stack = config.start_py3o
        _logger.debug("No command line start py3o %d", start_py3o_stack)

    if nmspc.start_rednerd:
        _logger.debug("Command line start rednerd %d", nmspc.start_rednerd)
        start_rednerd_stack = nmspc.start_rednerd == 1
    else:
        start_rednerd_stack = config.start_rednerd
        _logger.debug("No command line start rednerd %d", start_rednerd_stack)
    # TODO ajouter l’option de démarrer redis dans Configuration
    start_redis: bool = (nmspc.start_redis == 1) if nmspc.start_redis else False

    _logger.debug("Docker image: %s", image)
    # detect if docker image already exists
    client = DockerClient.client
    if not nmspc.no_build:
        failed = build_local_image(get_build_options(nmspc), nmspc.force_build)
        if failed:
            return failed

    # options is only used with subprocess call
    options: Dict[str, Any] = {
        "name": project_name,
        "tty": True,
        "extra_hosts": {},
        "ports": {},
        "environment": {},
    }
    if activate_s3_storage:
        options["environment"].update(
            {
                "s3AttachmentStorageActivate": "true",
                "s3AttachmentStorageServerBucket": activate_s3_storage,
            }
        )
    options["hostname"] = project_name
    # TODO detect that from configuration or force it
    exposed_port = 8069
    # TODO detect that from configuration or force it
    odoo_run_port = 8069
    if not gevent and not use_host_network:
        options["ports"][exposed_port] = odoo_run_port
    mounts: List[Mount] = []
    """Mounts for Odoo docker"""
    tested_modules: str = ""
    # TODO handle other version of odoo (base image need to be changed too)
    coverage = (
        (nmspc.test or nmspc.test_default or nmspc.install_test) and not test_tags
    ) and odoo_type not in (ODOO_7, ODOO_8, ODOO_9, ODOO_10)
    arg = [
        "populate"
        if populate_model or populate_size
        else ("coverage-start" if coverage else "start"),
        "--workers",
        str(workers),
    ]
    if db_password:
        arg.extend(("--db_password", db_password))
    if nmspc.update:
        arg.extend(("-u", nmspc.update, "--i18n-overwrite"))
        installing_or_updating = True
    if nmspc.test or nmspc.test_default or nmspc.install_test:
        arg.append("--test-enable")
    if nmspc.test:
        tested_modules = nmspc.test
        arg.extend(("-u", nmspc.test))
        if odoo_type == ODOO_7:
            arg.append("--log-level=test")
        installing_or_updating = True
    if nmspc.install_test:
        tested_modules = nmspc.install_test
        arg.extend(("-i", nmspc.install_test))
        if odoo_type == "odoo7":
            arg.append("--log-level=test")
        installing_or_updating = True
    if nmspc.test_default:
        test_modules = config.module_list_tests
        tested_modules = ",".join(test_modules)
        arg.extend(("-u", tested_modules))
        if odoo_type == ODOO_7:
            arg.append("--log-level=test")
        arg.append("--stop-after-init")
        installing_or_updating = True
    if tested_modules or nmspc.stop_after_init:
        arg.append("--stop-after-init")
    if database:
        arg.extend(("-d", database))
    if nmspc.install or nmspc.install_default:
        installing_or_updating = True
        modules_to_install = []
        if nmspc.install:
            modules_to_install.extend(nmspc.install.split(","))
        if nmspc.install_default:
            modules_to_install.extend(module_list)
        if modules_to_install:
            arg.append("-i")
            arg.append(",".join(modules_to_install))
    if nmspc.without_demo is not None:
        arg.append("--without-demo")
        arg.append(nmspc.without_demo)
    if nmspc.max_cron_threads:
        arg.append("--max-cron-threads=%s" % nmspc.max_cron_threads)
    if nmspc.log_level:
        arg.append("--log-level")
        arg.append(nmspc.log_level)
    if nmspc.log_handler:
        for lh in nmspc.log_handler + config.log_handler:
            arg.append("--log-handler")
            arg.append(lh)
    if nmspc.LOAD_LANGUAGE:
        load_language = nmspc.LOAD_LANGUAGE
    if load_language and installing_or_updating:
        arg.append("--load-language")
        arg.append(load_language)
    if (populate_model or populate_size) and odoo_type in (
        ODOO_7,
        ODOO_8,
        ODOO_9,
        ODOO_10,
        ODOO_11,
        ODOO_12,
        ODOO_13,
    ):
        _logger.warning("Populate is not available in your version")
    if populate_model:
        arg.append("--model")
        arg.append(populate_model)
    if populate_size:
        arg.append("--size")
        arg.append(populate_size)

    # auto detect local ip
    if use_host_network:
        local_ip = "127.0.0.1"
        options["network_mode"] = "host"
    else:
        local_ip = None
        try:
            from netifaces import AF_INET, ifaddresses  # type: ignore[import]

            iface_name = "docker0"
            addresses = [
                i["addr"]
                for i in ifaddresses(iface_name).setdefault(
                    AF_INET, [{"addr": "No IP addr"}]
                )
            ]
            if addresses:
                local_ip = addresses[0]
        except ImportError:
            _logger.warning(
                "Consider installing python netifaces to ease local IP detection"
            )
        if not local_ip:
            _logger.info("Contacting Google Public DNS to find our IP")
            for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]:
                s.connect(("8.8.8.8", 53))
                local_ip = s.getsockname()[0]
                s.close()
                break

        _logger.debug("IP found %s", local_ip)

    # auto detect local conf
    local_conf_path = nmspc.config
    user = db_user
    password = db_password
    if os.path.isfile(local_conf_path):
        _logger.info("Local configuration file found: %s", local_conf_path)
        if odoo_type in (ODOO_7, ODOO_8, ODOO_9, ODOO_10):
            odoo_conf_file = "/opt/odoo/etc/odoo.conf"
        else:
            odoo_conf_file = "/etc/opt/odoo/odoo.conf"
            # change ODOO_RC for version<ODOO_16
            options["environment"]["ODOO_RC"] = odoo_conf_file

        mounts.append(
            Mount(
                odoo_conf_file,
                os.path.join(project_path, local_conf_path),
                "bind",
            )
        )
        cp_local = ConfigParser()
        cp_local.read(local_conf_path)
        if not user:
            user = cp_local.get("options", "db_user", fallback=None)
        if not password:
            password = cp_local.get("options", "db_password", fallback=None)
        # doing this avoid having to change the value of redis_host
        # used by smile_redis_session_store
        redis_host = (
            cp_local.has_section("options")
            and cp_local.has_option("options", "redis_host")
            and cp_local.get("options", "redis_host")
        ) or None
        if redis_host:
            options["extra_hosts"][redis_host] = local_ip
    else:
        _logger.info("No configuration file at: %s", local_conf_path)
    # default values if nothing else
    if not user:
        # Odoo uses default if nothing is set (including in PGUSER env var)
        if start_postgresql:
            _logger.info("No user defined, using odoo when starting postgres")
            user = "odoo"
        else:
            # use environment variable
            user = os.getenv("PGUSER")
            # defaulting to odoo
            if not user:
                user = "odoo"
    arg.append("--db_user")
    arg.append(user)

    # directories to run chown on
    chown_directories = set()

    # data volume handling
    base_data_volume_name = "{}_data".format(project_name)
    if odoo_type == ODOO_7:
        # this path is in odoo code, the last part is the system user
        session_volume_name = "{}{}".format(base_data_volume_name, "_sessions")
        target = "/tmp/oe-sessions-{}".format("odoo")
        client.create_volume(session_volume_name, {"mounted in": target})
        mounts.append(Mount(target, session_volume_name))
        # by default odoo store attachments in database but it can be set on
        # disk. mount the volume in all cases.
        # XXX data is the path set in the database (in the form file:///data)
        #  it should be read from there, in that case target directory needs
        #  to include the database name.
        #  our databases only use file:///data so this should be fine
        target = (
            "/usr/local/lib/python2.7/dist-packages/"
            "openerp-7.0-py2.7.egg/openerp/data"
        )
        filestore_volume_name = "{}_filestore".format(base_data_volume_name)
        client.create_volume(filestore_volume_name, {"mounted in": target})
        mounts.append(Mount(target, filestore_volume_name))
    else:
        data_volume_path = "/var/lib/data"
        mounts.append(Mount(data_volume_path, base_data_volume_name))
        _, created = DockerClient.create_volume(
            base_data_volume_name,
            {
                "mounted in": data_volume_path,
            },
        )
        if created:
            chown_directories.add(data_volume_path)
        arg.append("--data-dir")
        arg.append(data_volume_path)
        if run_chown:
            chown_directories.add(data_volume_path)

    # avoid the duplication of unbind volumes with all addons
    if odoo_type == ODOO_7:
        extra_volumes = [
            "/opt/odoo/additional_addons",
            "/opt/odoo/var",
            # not used but declared as a mount point in Dockerfile :-(
            "/opt/odoo/data",
        ]
    elif odoo_type == ODOO_8:
        extra_volumes = [
            "/opt/odoo/additional_addons",
            "/opt/odoo/var",
            "/opt/odoo/data",
        ]
    elif odoo_type == ODOO_10:
        extra_volumes = [
            "/opt/odoo/additional_addons",
            "/opt/odoo/var",
            "/opt/odoo/data",
        ]
    elif odoo_type in (ODOO_11, ODOO_12, ODOO_13, ODOO_14):
        extra_volumes = ["/var/lib/odoo"]
    else:
        extra_volumes = ["/var/opt/odoo"]
    for extra_volume in extra_volumes:
        volume_name = "{}_{}".format(project_name, extra_volume.replace("/", "_"))
        _volume, created = DockerClient.create_volume(
            volume_name,
            {
                "mounted in": extra_volume,
                "comment": "only used to avoid creating too many volumes",
            },
        )
        mounts.append(Mount(extra_volume, volume_name))
        if created:
            chown_directories.add(extra_volume)

    # make sure the permission in the volumes are correct
    if run_chown:
        for extra_volume in extra_volumes:
            chown_directories.add(extra_volume)
        chown_directories.add("/mnt")

    if chown_directories:
        _logger.info("chown container’s %s to Odoo user", ", ".join(chown_directories))
        client.containers.run(
            image,
            command=["odoo", "--recursive"] + list(chown_directories),
            remove=True,
            entrypoint="/bin/chown",
            mounts=mounts,
            user="root",
        )

    if start_py3o_stack:
        start_py3o()
        # provide a couple of handy dns name for py3o
        options["extra_hosts"]["py3o"] = local_ip
        options["extra_hosts"]["py3o-fusion-server"] = local_ip

    # Database connection

    dbport = None
    if nmspc.dbport:
        dbport = str(nmspc.dbport)
        arg.append("--db_port")
        arg.append(dbport)

    db_host = nmspc.db_host

    create_database = False
    # Check that connection can be done, try to create user if asked to
    # TODO: handle the case where the database is still starting up
    # TODO: (and remove the sleep)
    if not odoo_help:
        if start_postgresql:
            if db_host:
                _logger.warning("Ignoring given db_host when starting postgres")
            if dbport:
                _logger.warning("Ignoring given db_port when starting postgres")
            db_host = "/var/run/postgresql"
            dbport = "5432"
            # Use socket rather than export port (avoid having to find a
            # free port number)
            pg, _stop_postgresql, socket_path = docker_run_postgresql(
                project_name, postgresql_version, host_pg_port=None
            )
            mounts.append(Mount(db_host, socket_path, "bind"))
        _logger.info("Testing database connection on base postgres")
        try:
            if start_postgresql:
                connect(
                    user=user,
                    password=password,
                    database="postgres",
                    host=socket_path,
                    port=dbport,
                )
                _logger.info("Connection successful")
            else:
                # psql uses PGHOST if provided and a socket otherwise
                if db_host:
                    pass
                elif "PGHOST" in os.environ:
                    db_host = os.getenv("PGHOST")
                    _logger.debug("Using PGHOST=%s", db_host)
                else:
                    # use socket, trying to guess it
                    # db_host = "/var/run/postgresql"
                    # _logger.debug("Defaulting to host %s", db_host)
                    db_host = local_ip
                if db_host.startswith("/"):
                    # mount the path inside the docker
                    mounts.append(Mount(db_host, db_host, "bind"))
                connect(
                    user=user,
                    password=password,
                    database="postgres",
                    host=db_host,
                    port=dbport,
                )

        except OperationalError as exception:
            if nmspc.create_user:
                _logger.debug("Cannot connect to database with user %s", user)
                _logger.debug(exception)
                _logger.info("Creating user %s", user)
                if start_postgresql:
                    connection = connect(
                        user="postgres",
                        database="postgres",
                        host=db_host,
                        port=5432,
                        password=POSTGRES_PASSWORD,
                    )
                else:
                    loginname = pwd.getpwuid(os.getuid())[0]
                    # use socket to create user
                    # TODO test if removing more arguments works
                    connection = connect(user=loginname, database="postgres")
                with connection.cursor() as cursor:
                    # not injection safe but you are on your own machine
                    # with already full access to db
                    cursor.execute(
                        "CREATE ROLE %s LOGIN CREATEDB PASSWORD %%s" % user,
                        (password,),
                    )
                connection.commit()
                connection.close()
            else:
                _logger.critical("Cannot connect to database with user %s", user)
                _logger.critical(exception)
                _logger.info(
                    "You can add the --create-user argument to try to create it"
                )
                return 16

        arg.append("--db_host")
        arg.append(db_host)

        # Create database if it does not exists

        if database:
            with connect(
                user=user,
                password=password,
                database="postgres",
                host=socket_path if start_postgresql else db_host,
                port=dbport,
            ) as connection:
                with connection.cursor() as cursor:
                    cursor.execute(
                        "SELECT count(*) FROM pg_database WHERE datname=%s", (database,)
                    )
                    data = cursor.fetchall()
                    create_database = data[0][0] == 0

    # --- restore / create database ---
    # TODO find out why odoo_help would stop us from restoring
    if start_postgresql and not odoo_help and (restore_filename or create_database):
        _logger.info("Creating database %s for %s", database, user)
        pg.exec_run(["createdb", "-U", user, database])
    if start_postgresql and not odoo_help and restore_filename:
        restore_basename = os.path.basename(restore_filename)
        _logger.info("Copying dump file in docker")
        inside_restore_filename = "/tmp/{}".format(restore_basename)
        DockerClient.put_file(restore_filename, pg, inside_restore_filename)
        _logger.info("Restoring database %s", database)
        restore, _output = pg.exec_run(
            [
                "pg_restore",
                "-U",
                user,
                "--no-owner",
                "--no-acl",
                "--job=4",
                "--dbname",
                database,
                inside_restore_filename,
            ]
        )
        if not restore:
            return 15
        _logger.info("Removing dump file in docker")
        pg.exec_run(["rm", inside_restore_filename])

    if not start_postgresql and not odoo_help and (restore_filename or create_database):
        connection = connect(
            user=user,
            password=password,
            database="postgres",
            host=db_host,
            port=dbport,
        )
        connection.autocommit = True
        _logger.info("Creating database %s", database)
        with connection.cursor() as cursor:
            cursor.execute(f'CREATE DATABASE "{database}"')
        connection.close()
    if not start_postgresql and not odoo_help and restore_filename:
        _logger.info("Restoring database %s", database)
        env = {
            "PGHOST": db_host,
            "PGPASSWORD": password,
            "PGUSER": user,
            "PG_COLOR": "always",
        }
        if dbport is not None:
            env["PGPORT"] = dbport
        restore = call(
            [
                "pg_restore",
                "--no-owner",
                "--no-acl",
                "--job=4",
                "--dbname",
                database,
                restore_filename,
            ],
            env=env,
        )
        if restore != 0:
            _logger.fatal("Errors during restoration")
            return 15

    # --- Pre SQL script ---
    if pre_sql_script:
        if start_postgresql:
            pre_sql_script_basename = os.path.basename(pre_sql_script)
            _logger.info("Copying SQL script %s in docker", pre_sql_script_basename)
            inside_pre_sql_script = "/tmp/{}".format(pre_sql_script_basename)
            client.put_file(pre_sql_script, pg, inside_pre_sql_script)
            _logger.info(
                "Running SQL script %s in %s",
                pre_sql_script_basename,
                database,
            )
            exit_code, _output = pg.exec_run(
                [
                    "psql",
                    "-U",
                    user,
                    "-d",
                    database,
                    "-f",
                    inside_pre_sql_script,
                ]
            )
            if exit_code > 0:
                _logger.fatal("Error when running SQL script %s", pre_sql_script)
                return 16
            _logger.info("Removing SQL script %s in docker", pre_sql_script)
            # remove script
            pg.exec_run(["rm", inside_pre_sql_script])
        else:
            _logger.info("Running SQL script %s on %s", pre_sql_script, database)
            env = {
                "PGHOST": db_host,
                "PGPASSWORD": password,
                "PGUSER": user,
                "PG_COLOR": "always",
            }
            if dbport is not None:
                env["PGPORT"] = dbport
            script = call(["psql", "--dbname", database, "-f", pre_sql_script], env=env)
            if script > 0:
                _logger.fatal("Error when running SQL script %s", script)
                return 16

    # --- Rednerd start ---

    if start_rednerd_stack:
        redner_db_user = nmspc.redner_db_user
        redner_db_password = nmspc.redner_db_password
        redner_admin_password = nmspc.redner_admin_password
        redner_user = nmspc.redner_user
        redner_password = nmspc.redner_password
        # share redner database by project rather than by database
        redner_database = f"redner_{project_name}"
        connect_kwargs: Dict[str, Any] = {
            "database": "postgres",
        }
        if start_postgresql:
            # uses the unix socket
            connect_kwargs.update(
                dict(
                    user="postgres",
                    host=db_host,
                    port="5432",
                    password=POSTGRES_PASSWORD,
                )
            )
        else:
            loginname = pwd.getpwuid(os.getuid())[0]
            # use socket to create user
            connect_kwargs.update(dict(user=loginname))
        with connect(**connect_kwargs) as connection:
            connection.autocommit = True
            with connection.cursor() as cursor:
                if nmspc.create_user:
                    try:
                        cursor.execute(
                            "CREATE ROLE %s LOGIN PASSWORD %%s" % redner_db_user,
                            (redner_db_password,),
                        )
                        _logger.info("Created database role %s", redner_db_user)
                    except DuplicateObject:
                        _logger.debug("Database role exists")
                try:
                    _logger.info("Creating database %s", redner_database)
                    cursor.execute(f'CREATE DATABASE "{redner_database}"')
                    _logger.info("Database created")
                except DuplicateDatabase:
                    # ignore exception if the database already exists
                    _logger.info("Database already exists")
        connect_kwargs["database"] = redner_database
        # do that here as we have superuser rights, while redner user probably does not
        with connect(**connect_kwargs) as connection:
            with connection.cursor() as cursor:
                _logger.info("Creating extension hstore if necessary")
                cursor.execute("CREATE EXTENSION IF NOT EXISTS hstore")
        if config.rednerd_repository is None:
            _logger.fatal(
                "Missing rednerd image name (rednerd_repository) in configuration"
            )
            return 81
        redner_docker = DockerRednerd(
            DockerClient.client,
            project_name,
            config.rednerd_repository,
            db_user=redner_db_user,
            db_password=redner_db_password,
            database=redner_database,
            db_host=db_host,
            db_port=dbport,
            attach=False,
            redner_user=redner_user,
            redner_password=redner_password,
            redner_admin_password=redner_admin_password,
            use_socket=True,
            version=config.rednerd_version,
        )
        redner_docker.start()
        # only useful when using port rather than socket. Not used at the moment
        options["extra_hosts"]["redner"] = local_ip
        # bind the socket dir in Odoo
        if hasattr(redner_docker, "socket_dir"):
            mounts.append(
                Mount(redner_docker.socket_dir, redner_docker.socket_dir, type="bind")
            )

        # --- Rednerd change config parameters in Odoo database ---
        if database:
            server_url = None
            # use unix socket if there is one, port otherwise.
            if hasattr(redner_docker, "socket_path"):
                server_url = redner_docker.socket_path
            elif hasattr(redner_docker, "redner_port"):
                server_url = f"http://localhost:{redner_docker.redner_port}/api/"
            _logger.info(
                "Changing redner ir.config_parameters in database %s", database
            )
            # connect to odoo database and set the ir.config_parameters
            with connect(
                user=user,
                password=password,
                database=database,
                host=db_host,
                port=dbport,
            ) as connection:
                with connection.cursor() as cursor:
                    set_rednerd_configparameters(
                        cursor, redner_user, redner_docker.api_key, server_url
                    )

    # --- Redis start ---

    redis_docker: Optional[DockerRedis] = None
    if start_redis:
        # TODO stop at exit à vrai sinon s’éteint pas?
        redis_docker = DockerRedis(DockerClient.client, project_name, attach=False)
        redis_docker.start()
        # bind the socket dir in Odoo
        mounts.append(
            Mount(redis_docker.socket_dir, redis_docker.socket_dir, type="bind")
        )
        options["environment"].update(
            # Variables used by flanker.
            REDIS_PORT="0",
            REDIS_DB="0",
            # Rely on a flanker patch to work.
            REDIS_UNIX_SOCKET_PATH=redis_docker.socket_path,
            # Variables used by session_redis
            ODOO_SESSION_REDIS="1",
            ODOO_SESSION_REDIS_URL=f"unix://{redis_docker.socket_path}?db=1",
        )

    # volume magic
    mount_list, _ = modules_mount(config)
    if mount_list:
        mounts.extend(mount_list)
    if odoo_type in (ODOO_7, ODOO_8):
        arg.append("--addons-path=/opt/odoo/sources/odoo/addons")
    # No way to set addons path to nothing for other versions unfortunately

    apply_flake8(nmspc, odoo_type)
    apply_isort(nmspc, config)
    apply_pylint(nmspc)

    # developer mode options
    if dev:
        if odoo_type in (ODOO_10, ODOO_11, ODOO_12, ODOO_13, ODOO_14, ODOO_15, ODOO_16):
            # automatically add reload if not already asks for
            if not any(opt in dev_opts for opt in ("all", "reload")):
                dev_opts.append("reload")
            # XXX pdb should not be run if not interactive
            # add pdb if no pubd, wdb, ipdb, pdb or all
            # the list of available options differs by odoo version and can be
            # seen with odoo --help, available with --odoo-help
            if not any(
                debugger in dev_opts
                for debugger in ("pubd", "wdb", "ipdb", "pdb", "all")
            ):
                dev_opts.append("pdb")
            # Add werkzeug if not already enabled
            if odoo_type not in (ODOO_10, ODOO_11, ODOO_12, ODOO_13, ODOO_14):
                if not any(opt in dev_opts for opt in ("all", "werkzeug")):
                    dev_opts.append("werkzeug")
            arg.append(f"--dev={','.join(set(dev_opts))}")
        if odoo_type in (ODOO_8,):
            arg.append("--auto-reload")
        if force_python_dev_mode or (
            config.pythondevmode and not force_no_python_dev_mode
        ):
            options["environment"]["PYTHONDEVMODE"] = 1

    if test_tags:
        arg.extend(("--test-tags", test_tags))

    project_version = "local"

    if os.path.exists(".hg"):
        # Next version guessing could be done by using marabunta yaml and using it
        # rather than doing latesttag.1
        # TODO change to use hglib
        project_version = check_output(
            [
                "hg",
                "identify",
                "-T",
                "{latesttag}{if('{dirty}', '.1.dev{latesttagdistance}{if('{topic}', ' {"
                "fqbn}')}', '{ifeq('{latesttagdistance}', 1, '', '{ifeq('{latesttagdist"
                "ance}', 2, '', '.1.dev{latesttagdistance}{if('{topic}', ' {fqbn}')}')}"
                "')}')}",
            ]
        ).decode()

    if os.path.exists(".git"):
        project_version = check_output(["git", "describe"]).decode()

    options["environment"]["VERSION_PACKAGE"] = ""
    options["environment"]["SENTRY_RELEASE"] = project_version
    options["environment"]["VERSION"] = project_version

    if odoo_help:
        arg.append("--help")

    # bind Odoo sources
    if odoo_sources:
        _logger.info("Binding Odoo sources at %s", odoo_sources)
        mounts.extend(odoo_source_mounts(config, odoo_sources))

    # avoid chowning at each run, faster but can cause trouble
    # newer odoo images does not use this environment variable anymore
    options["environment"]["ODOO_CHOWN"] = "false"

    if coverage:
        source_files = ",".join(
            [f"odoo.addons.{module}" for module in tested_modules.split(",")]
        )
        options["environment"]["COVERAGE_RUN_OPTIONS"] = (
            "--source=" + source_files + " --omit=*/__manifest__.py --branch"
        )

    if marabunta:
        marabunta_migration_file = config.marabunta_migration_file
        # if nothing is set, set the value to marabunta.yaml if there is such a file
        if not marabunta_migration_file:
            if os.path.exists("marabunta.yaml"):
                marabunta_migration_file = "marabunta.yaml"
    else:
        marabunta_migration_file = None

    if marabunta_migration_file is not None:
        marabunta_container_path = "/etc/marabunta.yaml"
        # add marabunta file to mounts
        mounts.append(
            Mount(
                marabunta_container_path,
                os.path.join(project_path, marabunta_migration_file),
                "bind",
                read_only=True,
            )
        )
        mounts.extend(anthem_mounts(config))

        options["environment"].update(
            {
                "MARABUNTA_MIGRATION_FILE": marabunta_container_path,
                "MARABUNTA_MODE": "local",
                "MARABUNTA_ALLOW_SERIE": "on",
            }
        )
        if start_postgresql:
            options["environment"]["MARABUNTA_DB_USER"] = (
                marabunta_db_user or "postgres"
            )
        elif marabunta_db_user:
            # only set if provided a value, otherwise the image will get the value from
            # command line arguments or the configuration file (using the same user os
            # the one used in odoo)
            options["environment"]["MARABUNTA_DB_USER"] = marabunta_db_user
        # marabunta is run inside the container by the start command
    else:
        # Only set to avoid problems with image that inherit another image using
        # marabunta while not using marabunta themselves
        options["environment"]["MARABUNTA_MIGRATION_FILE"] = ""
    if marabunta_force_version:
        options["environment"]["MARABUNTA_FORCE_VERSION"] = marabunta_force_version
    if marabunta_db_password:
        options["environment"]["MARABUNTA_DB_PASSWORD"] = marabunta_db_password

    if gevent:
        gevent_port = find_free_port()
        run_kwargs: Dict[str, Any] = {
            "command": ["caddy", "run", "--config", "/etc/caddy/Caddyfile"],
            "detach": True,
            "name": f"{project_name}_caddy",
            "mounts": [],
        }
        # force the port, so that the caddyfile is correct whatever the configuration
        if use_host_network:
            caddy_port = str(odoo_run_port)
            # change the port odoo runs on
            # TODO hopefully does not return 8069/odoo_run_port
            odoo_run_port = find_free_port()
            options["environment"]["MARABUNTA_WEB_PORT"] = str(odoo_run_port)
        else:
            caddy_port = "2019"
        # TODO pre odoo 13: check if http-port should be changed to xmlrpc-port
        gevent_uri = "/websocket"
        gevent_option = "--gevent-port"
        if odoo_type in (
            ODOO_7,
            ODOO_8,
            ODOO_9,
            ODOO_10,
            ODOO_11,
            ODOO_12,
            ODOO_13,
            ODOO_14,
            ODOO_15,
        ):
            gevent_uri = "/longpolling"
            gevent_option = "--longpolling-port"
        arg.extend(
            (
                gevent_option,
                str(gevent_port),
                "--http-port",
                str(odoo_run_port),
            )
        )

        odoo_host = "" if use_host_network else project_name
        caddyfile = f"""{{
    log {{
        format console
    }}
}}

:{caddy_port} {{
    reverse_proxy {gevent_uri}/* {odoo_host}:{gevent_port}
    reverse_proxy {odoo_host}:{odoo_run_port}
}}
"""
        with tempfile.NamedTemporaryFile(
            delete=False, prefix="Caddyfile."
        ) as caddyfile_tmp:
            caddyfile_tmp.write(caddyfile.encode("UTF-8"))

        def cleanup_caddyfile():
            os.unlink(caddyfile_tmp.name)

        atexit.register(cleanup_caddyfile)

        run_kwargs["mounts"].append(
            Mount("/etc/caddy/Caddyfile", caddyfile_tmp.name, type="bind")
        )
        if use_host_network:
            run_kwargs["network_mode"] = "host"
        else:
            try:
                client.networks.create(
                    name=project_name, driver="bridge", check_duplicate=True
                )
            except APIError:
                # TODO handle that more correctly
                # probably already exists
                pass
            run_kwargs.update(
                {
                    "network": project_name,
                    "ports": {caddy_port: odoo_run_port},
                    "hostname": "caddy",
                }
            )
            options["network"] = project_name

        _logger.info(
            "Starting caddy, use http://localhost:%s/ to access Odoo", caddy_port
        )
        # TODO there is an error as Odoo is not started up yet when this is started
        #  so maybe start it after Odoo started? (only happens if odoo windows are
        #  opened)
        # TODO also error when Odoo is powered off
        DockerClient.run("caddy", "latest", run_kwargs)

    interactive = True
    if interactive:
        options["stdin_open"] = True

    odoo_container = client.containers.create(
        image, command=arg, mounts=mounts, **options
    )
    _logger.info("Starting Odoo")
    if interactive:
        dockerpty.start(client.api, odoo_container.id)
        # already stopped but this is the easiest way to get the return code
        result_dict = odoo_container.wait()
        result = result_dict["StatusCode"]
    else:
        # XXX untested
        #  that might need a way to pass any signal to the container
        #  probably prevent the caddy start bellow
        result = odoo_container.start()

    # only remove after getting the return code
    odoo_container.remove()
    # Stop redis (python does not exit as the log process is still running)
    if redis_docker:
        _logger.info("Stopping redis")
        redis_docker.stop()
    return result


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
