#!/usr/bin/env python3
"""Run a redis docker with volume based on project name.
"""
import argparse
import logging
import os
import sys
import tempfile
import time
from typing import List, Optional

from docker.types import Mount  # type: ignore[import]

from .docker_client import DockerClient, DockerService
from .parsing import apply, basic_parser

__version__ = "1.0.0"
__date__ = "2023-03-20"
__updated__ = "2023-03-22"

_logger = logging.getLogger(__name__)


class DockerRedis(DockerService):
    """Redis service"""

    def __init__(
        self,
        docker_client,
        project_name: str,
        attach: bool = True,
        stop_at_exit: bool = True,
    ):
        super().__init__(
            docker_client,
            f"redis_{project_name}",
            "redis",
            "latest",
            True,
            attach=attach,
            stop_at_exit=stop_at_exit,
        )
        self.socket_dir = tempfile.mkdtemp(prefix="odoo_scripts_redis_socket-")
        """Redis socket’s directory."""
        os.chmod(self.socket_dir, 0o777)
        self.socket_path = os.path.join(self.socket_dir, "redis.sock")

    def _create_param(self) -> dict:
        result = super()._create_param()
        target = "/data"
        result["mounts"] = [
            Mount(
                target,
                DockerClient.create_volume(
                    f"{self.name}{target.replace('/', '_')}", {"mounted in": target}
                )[0].name,
                "volume",
            ),
            Mount("/var/run/redis", self.socket_dir, "bind"),
        ]
        result["command"] = [
            "redis-server",
            "--save",
            "--unixsocket",
            "/var/run/redis/redis.sock",
            "--port",
            "0",
        ]
        return result

    def start(self) -> bool:
        result = super().start()
        # by default the socket does not have permissions that will make it easy to use
        # (by odoo for example). User is 999, so that does not help much either.
        # Let redis start (alternative would be to run a command that check that the
        # socket exists, but it needs to be done inside the container due to permission
        # issue)
        time.sleep(0.01)
        DockerClient.exec(
            self._container, ["chmod", "777", "/var/run/redis/redis.sock"], "redis"
        )
        return result

    def _start(self) -> None:
        super()._start()
        if os.path.exists("/proc/sys/vm/overcommit_memory"):
            with open("/proc/sys/vm/overcommit_memory", "r") as f:
                data = f.read()
                if data == "0\n":
                    _logger.info(
                        "The warning in redis about memory overcommit applies to the "
                        "host system"
                    )

    def __del__(self):
        os.rmdir(self.socket_dir)


def __parser() -> argparse.ArgumentParser:
    program_version_message = f"%(prog)s {__version__} ({__updated__})"
    program_short_description = __doc__.split(".")[0]
    program_license = f"""{program_short_description}

      Created by Vincent Hatakeyama on {__date__}.
      Copyright 2023 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """
    parser = basic_parser(program_license, program_version_message)
    project_path = os.path.realpath(".")
    project_name = os.path.basename(project_path)
    parser.add_argument(
        "--name",
        help="Container name to use [default: %(default)s]",
        default=project_name,
    )
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    parser = __parser()
    namespace = parser.parse_args(argv)
    apply(namespace)
    docker_redis = DockerRedis(DockerClient.client, namespace.name)
    docker_redis.start()
    return 0


if __name__ == "__main__":
    sys.exit(main())
