#!/usr/bin/env python3
"""This script update another repository that is a copy of the sources
"""
import logging
import os
import shutil
import sys
from subprocess import call
from typing import List, Optional

import hglib  # type: ignore[import]

from .config import Config
from .parsing import apply, basic_parser

_logger = logging.getLogger(__name__)

__version__ = "1.2.0"
__date__ = "2020-06-19"
__updated__ = "2023-02-27"


def __parser():
    program_version = __version__
    program_build_date = str(__updated__)
    program_version_message = "%%(prog)s %s (%s)" % (
        program_version,
        program_build_date,
    )
    program_shortdesc = __doc__.split(".", maxsplit=1)[0]
    program_license = """%s

      Created by Vincent Hatakeyama on %s.
      Copyright 2018, 2020, 2021, 2023 XCG Consulting. All rights reserved.

      Licensed under the MIT License

      Distributed on an "AS IS" basis without warranties
      or conditions of any kind, either express or implied.

    USAGE
    """ % (
        program_shortdesc,
        str(__date__),
    )
    parser = basic_parser(program_license, program_version_message)
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Copy modules for a build, callable version that parses arguments"""
    parser = __parser()
    nmspc = parser.parse_args(argv)
    apply(nmspc)
    # XXX probably needs an option to indicate path to use, defaulting to
    #  default
    _update_duplicate_sources()
    return 0


def _update_duplicate_sources(**kwargs):
    # detect mercurial/git. if git, fail.
    if os.path.exists(".git"):
        raise Exception("Git source duplication not handled for now")
    _update_duplicate_sources_mercurial(**kwargs)


def _update_duplicate_sources_mercurial(directory: str = ".", path: str = "default"):
    """
    :arg directory: path to the base directory
    :arg path: path to use (from `.hg/hgrc`)
    """
    mercurial = hglib.open(directory)
    # get branch and tags
    branch = mercurial.identify(branch=True).strip()
    _logger.info("Current branch %s", branch.decode())
    tags = mercurial.identify(tags=True).split()
    if b"tip" in tags:
        tags.remove(b"tip")
    _logger.info("Current tags %s", ", ".join(tag.decode() for tag in tags))
    # XXX get topics?
    # find the duplicate repository URL
    conf = Config()
    if conf.duplicate_repo:
        duplicate_path = conf.duplicate_repo.encode()
    else:
        remote_path = mercurial.paths(path.encode())
        duplicate_path = "{}_sources".format(remote_path.decode()).encode()
    _logger.info("Clone path %s", duplicate_path.decode())
    # clone the duplicate repository
    duplicate_destination = b"sources"
    if os.path.exists(duplicate_destination):
        raise Exception(
            "Destination directory '{}' exists!".format(duplicate_destination.decode())
        )
    _logger.info("Cloning in %s", duplicate_destination.decode())
    mercurial.clone(source=duplicate_path, dest=duplicate_destination)
    hg_duplicate = hglib.open(duplicate_destination)
    # update clone to same branch if it exists.
    branch_exists_in_duplicate = False
    for branch_name, num, id_ in hg_duplicate.branches():
        if branch_name == branch:
            hg_duplicate.update(rev=num)
            branch_exists_in_duplicate = True
            _logger.info(
                "Found branch %s in duplicate (%d, %s)",
                branch_name.decode(),
                num,
                id_,
            )
            break
    # TODO otherwise put ourselves after parent branch (that is not done in
    #  shell script)
    if not branch_exists_in_duplicate:
        hg_duplicate.update(b"tip")
        _logger.info(
            "Branch %s not found in duplicate, updating to tip", branch.decode()
        )
    # rsync content inside the other directory
    cmd = (
        [
            "rsync",
            "--delete",
            "--cvs-exclude",
            "--exclude=.hg*",
            "--exclude=.git*",
            "--exclude='*.pyc'",
            "--exclude=.drone.yml",
            "--exclude=.gitlab-ci.yml",
            "--copy-unsafe-links",
            "--recursive",
            "--relative",
            "--times",
        ]
        + conf.modules
        + conf.dependencies
        + conf.other_sources
        + list(conf.python_packages.keys())
        + [duplicate_destination.decode()]
    )
    _logger.debug(" ".join(cmd))
    call(cmd)
    # create the branch if it does not exist
    if not branch_exists_in_duplicate:
        hg_duplicate.branch(name=branch)
    # commit
    _logger.info("Copying files")
    hg_duplicate.commit(message=b"Update sources", addremove=True)
    # retag
    if tags:
        _logger.info("Applying tags")
        hg_duplicate.tag(tags)
    else:
        _logger.info("No tag to apply")
    # push
    _logger.info("Pushing duplicate")
    hg_duplicate.push(newbranch=not branch_exists_in_duplicate)
    # clean up cloned repository
    _logger.info("Cleaning up duplicate sources")
    shutil.rmtree(duplicate_destination)


if __name__ == "__main__":
    sys.exit(main())
